#version 430 core
/*
* Copyright 2018 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
	vec4 shadow_coord;
} vert;

out vec4 color;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;
uniform sampler2D diffuseTexture;
uniform sampler2DShadow shadowMap;
uniform samplerCube cubemap;
uniform float reflectionCoefficient = 0;
uniform float brightnessCorrection = 1;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;


vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation, float visibility) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return visibility * ((diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att); 
}

// Returns a random number based on a vec3 and an int.
float random(vec3 seed, int i){
	vec4 seed4 = vec4(seed,i);
	float dot_product = dot(seed4, vec4(12.9898,78.233,45.164,94.673));
	return fract(sin(dot_product) * 43758.5453);
}

float getVisibility(vec3 n, vec3 l) {
	float cosTheta = max(0, dot(n, -dirL.direction));
	
	//float bias = 0.002;
	float bias = 0.002*tan(acos(cosTheta));
	bias = max(bias, 0);
	bias = min(bias, 0.05);
	
	float visibility = 1.0;
	for (int i=0;i<16;i++){
		int index = i;
		//calculate visibility
		visibility -= 0.05*(1.0-texture( shadowMap, vec3(vert.shadow_coord.xy + poissonDisk[index]/700.0,  (vert.shadow_coord.z - bias)/vert.shadow_coord.w) ));
	}
	
	return visibility;
}

//for environment mapping
vec3 environment(vec3 r) {
	return vec3(texture(cubemap, r).rgb) * 0.5;
}

void main() {	
	vec3 n = normalize(vert.normal_world);
	vec3 v = normalize(camera_world - vert.position_world);
	vec3 r = reflect(-v, n);
	
	vec4 texColorAlpha = texture(diffuseTexture, vert.uv);
	vec3 texColor = texColorAlpha.rgb;
	color = vec4(texColor * materialCoefficients.x, texColorAlpha.a); // ambient


	//get visibility for shadow map
	float visibility = getVisibility(n, -dirL.direction);
	
	// add directional light contribution
	color.rgb += phong(n, -dirL.direction, v, dirL.color * texColor, materialCoefficients.y, dirL.color, materialCoefficients.z, specularAlpha, false, vec3(0), visibility);
			
	// add point light contribution
	color.rgb += phong(n, pointL.position - vert.position_world, v, pointL.color * texColor, materialCoefficients.y, pointL.color, materialCoefficients.z, specularAlpha, true, pointL.attenuation, visibility);
	
	color.rgb = color.rgb*(1 - reflectionCoefficient) + environment(r) * reflectionCoefficient;
	
	color.rgb *= brightnessCorrection;
	
	color.a = texColorAlpha.a;

}

